import math
from typing import Union
from pathlib import Path
from planar_magnetics.geometry import Arc, Polygon, Point
from planar_magnetics.smoothing import smooth_polygon
from planar_magnetics.utils import dcr_of_annulus


class Spiral:
    """Create an optimized spiral multi-turn winding on a single layer
    """

    def __init__(
        self,
        at: Point,
        inner_radius: float,
        outer_radius: float,
        num_turns: int,
        gap: float,
        layer: str = "F.Cu",
        radius: float = 0.1e-3,
    ):

        # calculate optimal turn radii using equation 10 from Conceptualization and Analysis of a
        # Next-Generation Ultra-Compact 1.5-kW PCB-Integrated Wide-Input-Voltage-Range 12V-Output
        # Industrial DC/DC Converter Module
        # https://www.pes-publications.ee.ethz.ch/uploads/tx_ethpublications/7_electronics-10-02158_FINAL_Knabben.pdf
        inverse_num_turns = 1 / num_turns
        radii = [
            (inner_radius ** (num_turns - i) * outer_radius ** i) ** inverse_num_turns
            for i in range(num_turns)
        ]

        # verify that the minimum trace width is greater than 2 x min radius
        min_trace_width = (
            radii[1] - radii[0] - gap if num_turns > 1 else outer_radius - inner_radius
        )
        assert (
            min_trace_width > 2 * radius
        ), f"This spiral requires a min trace width of {1e3*min_trace_width}mm, which is less than 2 x radius"

        # create the arcs for the inner turns
        angle = math.acos(1 - gap / radii[0])
        arcs = [Arc(at, radii[0], -math.pi + angle, math.pi)]
        for r0, r1 in zip(radii[0:-1], radii[1:]):
            angle = math.acos(gap / r1 + r0 * (1 - gap / r0) / r1)
            arc = Arc(at, r1, -math.pi + angle, math.pi)
            arcs.append(arc)

        # create the outermost arc
        a0 = math.acos(
            gap / outer_radius + radii[-1] * (1 - gap / radii[-1]) / outer_radius
        )
        a1 = math.acos(radii[-1] * (1 - gap / radii[-1]) / outer_radius)
        arc = Arc(at, outer_radius, math.pi + a0, -math.pi + a1)
        arcs.append(arc)

        # create the outer arcs of the other turns
        for r0, r1 in zip(radii[-2::-1], radii[-1:0:-1]):
            angle = math.acos((r0 - gap) / (r1 - gap))
            arc = Arc(at, r1 - gap, math.pi, -math.pi + angle)
            arcs.append(arc)

        polygon = Polygon(arcs, layer)

        # Add smoothing if a positive radius is provided
        if radius > 0:
            polygon = smooth_polygon(polygon, radius)

        self.polygon = polygon

        self.inner_radii = radii
        self.outer_radii = [r - gap for r in radii[1:]] + [outer_radius]

    def estimate_dcr(self, thickness: float, rho: float = 1.68e-8):
        """Estimate the DC resistance of the winding

        This function will estimate the DC resistance of the winding by calculating the estimated
        dc resistance of each turn and adding the estimated inter-turn via resistance 
        
        Args:
            thickness (float): The copper thickness of the layer
            rho (float): The conductivity of the material used in the layer

        Returns:
            float: An estimation of the DC resistance in ohms
        """

        # sum the resistance of each turn
        resistance = 0
        for r0, r1 in zip(self.inner_radii, self.outer_radii):
            resistance += dcr_of_annulus(thickness, r0, r1, rho)

        return resistance

    def plot(self, max_angle: float = math.pi / 36):
        self.polygon.plot(max_angle)

    def export_to_dxf(
        self,
        filename: Union[str, Path],
        version: str = "R2000",
        encoding: str = None,
        fmt: str = "asc",
    ) -> None:
        """Export the polygon to a dxf file

        Args:
            filename: file name as string
            version: DXF version
            encoding: override default encoding as Python encoding string like ``'utf-8'``
            fmt: ``'asc'`` for ASCII DXF (default) or ``'bin'`` for Binary DXF
        """

        self.polygon.export_to_dxf(filename, version, encoding, fmt)

    def __str__(self):
        """Print KiCAD S-Expression of the spiral.  Assumes units are mm."""
        return self.polygon.__str__()


if __name__ == "__main__":

    from planar_magnetics.creepage import calculate_creepage
    from planar_magnetics.utils import weight_to_thickness

    # create a spiral inductor
    spiral = Spiral(
        at=Point(110e-3, 110e-3),
        inner_radius=6e-3,
        outer_radius=12e-3,
        num_turns=1,
        gap=calculate_creepage(500, 1),
        layer="F.Cu",
        radius=0.3e-3,
    )

    # estimate the dc resistance of this spiral assuming 2 oz copper
    dcr = spiral.estimate_dcr(thickness=weight_to_thickness(2))
    print(f"Estimated DCR of this spiral is {dcr*1e3} mOhms")

    # dispay a preview of the spiral from Python using matplotlib
    spiral.plot()

    # # export this to a DXF file
    # spiral.export_to_dxf("spiral.dxf")

    # get the KiCad S expression, which can be then be copy-pasted into a KiCAD footprint file and edited from the footprint editer
    print(spiral)



import math
from planar_magnetics.geometry import Point
from planar_magnetics.cores import Core
from planar_magnetics.creepage import Classification, calculate_creepage
from planar_magnetics.kicad import Footprint, Pad, PadType, Reference, Value
from planar_magnetics.windings.single import TopTurn, InnerTurn, BottomTurn, ViaStrip


class Winding:
    def __init__(
        self,
        at: Point,
        inner_radius: float,
        outer_radius: float,
        number_layers: int,
        gap: float = 0.5e-3,
        termination_width: float = None,
        viastrip_width: float = 1e-3,
    ):

        self.number_layers = number_layers

        if termination_width is None:
            termination_width = outer_radius - inner_radius

        # calculate other useful angles
        inner_gap_angle = math.asin(gap / inner_radius)
        term_angle = math.asin(termination_width / outer_radius)

        # calculate the angle we can allocate to the via transitions
        circumfrance_for_transitions = (
            2 * math.pi - term_angle
        ) * inner_radius - number_layers * gap
        angle_for_transitions = circumfrance_for_transitions / inner_radius
        viastrip_angle = angle_for_transitions / (number_layers - 1)

        # calculate other useful angles
        inner_gap_angle = math.asin(gap / inner_radius)
        term_angle = math.asin(termination_width / outer_radius)

        # calculate the required rotation per turn
        initial_rotation = (term_angle + inner_gap_angle) / 2
        rotation_per_turn = viastrip_angle + inner_gap_angle

        # create the top and bottom turns
        top = TopTurn(
            at,
            inner_radius,
            outer_radius,
            gap,
            termination_width,
            viastrip_angle,
            viastrip_width,
            "F.Cu",
        )
        inners = [
            InnerTurn(
                at,
                inner_radius,
                outer_radius,
                gap,
                -n * rotation_per_turn - initial_rotation,
                viastrip_angle,
                viastrip_width,
                f"In{n}.Cu",
            )
            for n in range(1, number_layers - 1)
        ]
        bottom = BottomTurn(
            at,
            inner_radius,
            outer_radius,
            gap,
            termination_width,
            viastrip_angle,
            viastrip_width,
            "B.Cu",
        )
        self.turns = [top] + inners + [bottom]

        # create the via strips
        initial_angle = initial_rotation + inner_gap_angle / 2

        layers = [(t.layer, b.layer) for t, b in zip(self.turns[0:-1], self.turns[1:])]

        self.viastrips = [
            ViaStrip(
                at,
                layers[n],
                inner_radius,
                -initial_angle - n * rotation_per_turn,
                -initial_angle - n * rotation_per_turn - viastrip_angle,
                0.8e-3,
                0.4e-3,
            )
            for n in range(number_layers - 1)
        ]

    def estimate_dcr(self, thicknesses: [float], rho: float = 1.68e-8):
        """Estimate the DC resistance of the winding

        This function will estimate the DC resistance of the winding by calculating the estimated
        dc resistance of each turn and adding the estimated inter-turn via resistance 
        
        Args:
            thicknesses: The thickness of each layer in the winding
            rho (float): The conductivity of the material used in the layer

        Returns:
            float: An estimation of the DC resistance in ohms
        """

        assert (
            len(thicknesses) == self.number_layers
        ), f"You need to specify 1 thickness for each layer, so len(thicknesses) should be {self.number_layers}, not{len(thicknesses)}"

        resistance = 0
        for thickness, turn in zip(thicknesses, self.turns):
            resistance += turn.estimate_dcr(thickness, rho)

        return resistance

    def __str__(self):
        turns = "\n".join(turn.__str__() for turn in self.turns)
        vias = "\n".join(viastrip.__str__() for viastrip in self.viastrips)
        expression = turns + vias
        return expression


class Cffc:
    def __init__(
        self,
        inner_radius: float,
        outer_radius: float,
        number_turns: int,
        voltage: float,
        classification: Classification = Classification.B4,
    ):
        origin = Point(0, 0)

        self.inner_radius = inner_radius
        self.outer_radius = outer_radius

        # calculate the required creepage distance
        creapage = calculate_creepage(voltage, classification)

        self.termination_width = outer_radius - inner_radius

        self.number_turns = number_turns
        self.number_layers = number_turns + 1

        # create the windings
        self.winding = Winding(
            at=origin,
            inner_radius=inner_radius,
            outer_radius=outer_radius,
            number_layers=self.number_layers,
            gap=creapage,
            termination_width=self.termination_width,
        )

        # create the core
        self.core = Core(
            at=origin,
            inner_radius=inner_radius,
            outer_radius=outer_radius,
            termination_width=self.termination_width,
            edge_to_trace=0.635e-3,
            edge_to_core=0.5e-3,
        )

    def __str__(self):
        expression = self.winding.__str__() + self.core.__str__()
        return expression

    def estimate_dcr(self, thicknesses: [float], rho: float = 1.68e-8):
        """Estimate the DC resistance of the winding

        This function will estimate the DC resistance of the winding by calculating the estimated
        dc resistance of each turn and adding the estimated inter-turn via resistance 
        
        Args:
            thicknesses: The thickness of each layer in the winding
            rho (float): The conductivity of the material used in the layer

        Returns:
            float: An estimation of the DC resistance in ohms
        """

        return self.winding.estimate_dcr(thicknesses, rho)

    def to_kicad_footprint(self, name: str):
        """Export the Cffc inductor design as a KiCAD footprint file (*.kicad_mods)
        """

        # vias are not allowed in KiCAD footprints, so convert the via strips to through-hole pads
        pads = [
            via.to_pad() for viastrip in self.winding.viastrips for via in viastrip.vias
        ]

        # add the termination pads
        location = Point(self.outer_radius + self.termination_width / 2, 0)
        size = self.termination_width / 2
        pads.extend(
            [
                Pad(PadType.SMD, 1, location, size, ("F.Cu",)),
                Pad(PadType.SMD, 2, location, size, ("B.Cu",)),
            ]
        )

        # add the reference and value silkscreens
        x_loc = self.core.width / 2 + 1e-3
        height_avail = (self.core.width - self.termination_width) / 2
        font_size = min(2e-3, height_avail / 4)
        val_loc = Point(x_loc, self.termination_width / 2 + height_avail / 3)
        ref_loc = Point(x_loc, self.termination_width / 2 + 2 * height_avail / 3)
        reference = Reference(ref_loc, font_size)
        value = Value(val_loc, font_size)

        # create a footprint from the various elements
        contents = [self.core] + self.winding.turns + pads + [reference, value]
        footprint = Footprint(name, contents=contents)

        # write the footprint to a file
        fh = open(f"{name}.kicad_mod", "w")
        fh.write(footprint.__str__())
        fh.close()


if __name__ == "__main__":

    from planar_magnetics.utils import weight_to_thickness

    inductor = Cffc(inner_radius=4.9e-3, outer_radius=9e-3, number_turns=5, voltage=500)

    # estimate the dc resistance of this inductor
    # using the CFFC structure, a 5 turn inductor requires 6 layers
    # assume we are using 1.5 oz on top/botton and 2oz on interior layers
    thicknesses = [
        weight_to_thickness(1.5),
        weight_to_thickness(2),
        weight_to_thickness(2),
        weight_to_thickness(2),
        weight_to_thickness(2),
        weight_to_thickness(1.5),
    ]
    dcr = inductor.estimate_dcr(thicknesses)
    print(f"Estimated DCR of this inductor is {dcr*1e3} mOhms")

    # create a complete KiCAD footprint
    inductor.to_kicad_footprint("cffc_inductor")



from enum import Enum


class Classification(Enum):
    B1 = 1, "Internal Conductors"
    B2 = 2, "External Conductors, uncoated, sea level to 3050 m"
    B3 = 3, "External Conductors, uncoated, over 3050 m"
    B4 = 4, "External Conductors, with permanent polymer coating (any elevation)"
    A5 = 5, "External Conductors, with conformal coating over assembly (any elevation)"
    A6 = 6, "External Component lead/termination, uncoated"
    A7 = (
        7,
        "External Component lead termination, with conformal coating (any elevation)",
    )

    def __new__(cls, *args, **kwds):
        obj = object.__new__(cls)
        obj._value_ = args[0]
        return obj

    # ignore the first param since it's already set by __new__
    def __init__(self, _: str, description: str = None):
        self._description_ = description

    def __str__(self):
        return self.description

    # this makes sure that the description is read-only
    @property
    def description(self):
        return self._description_


# Creepage table
CREEPAGE_TABLE = {
    15: [0.05e-3, 0.1e-3, 0.1e-3, 0.05e-3, 0.13e-3, 0.13e-3, 0.13e-3],
    30: [0.05e-3, 0.1e-3, 0.1e-3, 0.05e-3, 0.13e-3, 0.25e-3, 0.13e-3],
    50: [0.1e-3, 0.6e-3, 0.6e-3, 0.13e-3, 0.13e-3, 0.4e-3, 0.13e-3],
    100: [0.1e-3, 0.6e-3, 1.5e-3, 0.13e-3, 0.13e-3, 0.5e-3, 0.13e-3],
    150: [0.2e-3, 0.6e-3, 3.2e-3, 0.4e-3, 0.4e-3, 0.8e-3, 0.4e-3],
    170: [0.2e-3, 1.25e-3, 3.2e-3, 0.4e-3, 0.4e-3, 0.8e-3, 0.4e-3],
    250: [0.2e-3, 1.25e-3, 6.4e-3, 0.4e-3, 0.4e-3, 0.8e-3, 0.4e-3],
    300: [0.2e-3, 1.25e-3, 12.5e-3, 0.4e-3, 0.4e-3, 0.8e-3, 0.4e-3],
    500: [0.25e-3, 2.5e-3, 12.5e-3, 0.8e-3, 0.8e-3, 1.5e-3, 0.8e-3],
}

PER_VOLT_TABLE = [
    0.0025e-3,
    0.005e-3,
    0.025e-3,
    0.00305e-3,
    0.00305e-3,
    0.00305e-3,
    0.00305e-3,
]


def calculate_creepage(voltage: float, classification: Classification):
    """Calculates the minimum creepage distance

    Args:
        voltage: RMS voltage

    Returns:
        float: The recommended creepage distance

    """

    # re-cast to classification to allow users to pass an integer as the classification
    classification = Classification(classification)

    for key, value in CREEPAGE_TABLE.items():
        if key >= voltage:
            return value[classification.value - 1]

    # if the voltage is >500V, the per/volt table must be added
    base_volt = CREEPAGE_TABLE[500][classification.value - 1]
    per_volt = PER_VOLT_TABLE[classification.value - 1]
    creepage = base_volt + (voltage - 500) * per_volt
    return creepage


if __name__ == "__main__":

    classification = Classification(3)
    print(classification)
    creepage = calculate_creepage(1000, 3)
    print(creepage)



import math
import uuid
from planar_magnetics.geometry import (
    Arc,
    Point,
    Polygon,
    TWO_PI,
    PI_OVER_TWO,
)


def calculate_core_extension(area: float, radius: float, opening_width: float) -> float:
    """Calculate the required core extension to have the same area as the center-post

    Args:
        area (float): The total required area of the outer legs
        radius (float): The radius of the edge of the outer legs

    Returns:
        float: The required extension
    """

    # first, calculate the area of a single leg with no extension
    start_angle = math.asin((opening_width / 2) / radius)
    end_angle = PI_OVER_TWO - start_angle

    x = math.sqrt(radius ** 2 - (opening_width / 2) ** 2)

    # start with the area of the square from the center to the corner
    min_leg_area = x ** 2

    # subtract the opening triangles
    min_leg_area -= x * opening_width / 2

    # subtract the arc
    min_leg_area -= ((end_angle - start_angle) / TWO_PI) * (math.pi * radius ** 2)

    # calculated the required extension area
    extension_area = area / 4 - min_leg_area

    # if the minimum leg area is already sufficient, then just return 0
    if extension_area <= 0:
        return 0

    # calculate the minimum width of the leg
    x = radius * (math.cos(start_angle) - math.cos(end_angle))

    # solve for the extension using the quadratic equasion
    extension = (-2 * x + math.sqrt(4 * x ** 2 + 4 * extension_area)) / 2

    return extension


class Core:
    def __init__(
        self,
        at: Point,
        inner_radius: float,
        outer_radius: float,
        termination_width: float = None,
        edge_to_trace: float = 0.635e-3,
        edge_to_core: float = 0.5e-3,
    ):

        if termination_width is None:
            termination_width = outer_radius - inner_radius

        self.at = at
        self.inner_radius = inner_radius
        self.outer_radius = outer_radius
        self.edge_to_trace = edge_to_trace
        self.edge_to_core = edge_to_core
        self.tstamp = uuid.uuid4()

        # calculate core dimensions
        self.centerpost_radius = inner_radius - edge_to_trace - edge_to_core
        self.outerpost_radius = outer_radius + edge_to_trace + edge_to_core

        # calculate the centerpost area
        self.centerpost_area = math.pi * self.centerpost_radius ** 2

        # calculate the radius of the outer post cutouts
        outer_cutout_radius = self.outerpost_radius - edge_to_core

        # calculate the start angle first outer post cutout leg
        start_angle = math.asin(
            (termination_width / 2 + edge_to_trace) / outer_cutout_radius
        )

        # calculate the end angle of the first outer post cutout leg
        end_angle = math.pi / 2 - start_angle

        # create polygons for the outer post cutouts
        extension = calculate_core_extension(
            area=self.centerpost_area,
            radius=self.outerpost_radius,
            opening_width=termination_width + 2 * edge_to_trace + 2 * edge_to_core,
        )

        cutout_extension = extension + edge_to_core

        # calculate the arcs
        arc = Arc(at, outer_cutout_radius, start_angle, end_angle)
        corner1 = arc.end + Point(0, cutout_extension)
        corner3 = arc.start + Point(cutout_extension, 0)
        corner2 = Point(corner3.x, corner1.y)
        leg1 = Polygon([arc, corner1, corner2, corner3], "Edge.Cuts", 0.1e-3, "none")

        # calculate the core cutout dimensions, which can be used for silkscreen placement
        self.width = 2 * (arc.start.x + cutout_extension)
        self.length = self.width

        start_angle += math.pi / 2
        end_angle += math.pi / 2
        arc = Arc(at, outer_cutout_radius, start_angle, end_angle)
        corner1 = arc.end + Point(-cutout_extension, 0)
        corner3 = arc.start + Point(0, cutout_extension)
        corner2 = Point(corner1.x, corner3.y)
        leg2 = Polygon([arc, corner1, corner2, corner3], "Edge.Cuts", 0.1e-3, "none")

        start_angle += math.pi / 2
        end_angle += math.pi / 2
        arc = Arc(at, outer_cutout_radius, start_angle, end_angle)
        corner1 = arc.end + Point(0, -cutout_extension)
        corner3 = arc.start + Point(-cutout_extension, 0)
        corner2 = Point(corner3.x, corner1.y)
        leg3 = Polygon([arc, corner1, corner2, corner3], "Edge.Cuts", 0.1e-3, "none")

        start_angle += math.pi / 2
        end_angle += math.pi / 2
        arc = Arc(at, outer_cutout_radius, start_angle, end_angle)
        corner1 = arc.end + Point(cutout_extension, 0)
        corner3 = arc.start + Point(0, -cutout_extension)
        corner2 = Point(corner1.x, corner2.y)
        leg4 = Polygon([arc, corner1, corner2, corner3], "Edge.Cuts", 0.1e-3, "none")

        self.outerposts = [leg1, leg2, leg3, leg4]

    def __str__(self):

        # create the centerpost milling
        end = self.at + Point(self.centerpost_radius + self.edge_to_core, 0)
        layer = "Edge.Cuts"
        centerpost = f"(fp_circle (center {self.at}) (end {end}) (layer {layer}) (width 0.1) (fill none) (tstamp {self.tstamp}))"

        # create the milling for each corner
        outerposts = "\n".join(outerpost.__str__() for outerpost in self.outerposts)

        expression = "\n".join([centerpost, outerposts])

        return expression



import math
from planar_magnetics.geometry import Arc, Point, Polygon
from planar_magnetics.kicad import Via
from planar_magnetics.utils import dcr_of_annulus


class TopTurn:
    """Defines a top layer turn of a CFFC inductor
    """

    def __init__(
        self,
        at,
        inner_radius: float,
        outer_radius: float,
        gap: float,
        termination_width: float,
        viastrip_angle: float,
        viastrip_width: float,
        layer: str,
    ):

        self.inner_radius = inner_radius
        self.outer_radius = outer_radius
        self.layer = layer

        # calculate the gap angles
        inner_gap_angle = math.asin(gap / inner_radius)
        outer_gap_angle = math.asin(gap / outer_radius)

        # angle from "at" to the corner of the termination
        term_angle = math.asin(termination_width / outer_radius / 2)

        termination_arc = Arc(at, inner_radius, -term_angle, term_angle)

        # create the inner arc
        inner_arc = Arc(
            at,
            inner_radius + viastrip_width,
            term_angle,
            2 * math.pi - term_angle - inner_gap_angle - viastrip_angle,
        )

        via_arc = Arc(
            at,
            inner_radius,
            2 * math.pi - term_angle - inner_gap_angle - viastrip_angle,
            2 * math.pi - term_angle - inner_gap_angle,
        )

        # create outer arc
        outer_arc = Arc(
            at, outer_radius, 2 * math.pi - term_angle - outer_gap_angle, term_angle,
        )

        # create termination
        termination = [
            at + Point(outer_radius + termination_width, termination_width / 2),
            at + Point(outer_radius + termination_width, -termination_width / 2),
            at + Point(outer_radius * math.cos(term_angle), -termination_width / 2),
        ]

        # create the polygon
        points = [termination_arc, inner_arc, via_arc, outer_arc] + termination
        self.polygon = Polygon(points, layer)

    def estimate_dcr(self, thickness: float, rho: float = 1.68e-8):
        """Estimate the DC resistance of the winding

        This function will estimate the DC resistance of the winding by calculating the estimated
        dc resistance of each turn and adding the estimated inter-turn via resistance 
        
        Args:
            thickness (float): The copper thickness of the layer
            rho (float): The conductivity of the material used in the layer

        Returns:
            float: An estimation of the DC resistance in ohms
        """

        # estimate the resistance of the turn
        turn_resistance = dcr_of_annulus(
            thickness, self.inner_radius, self.outer_radius, rho
        )

        # TODO:  Need to add via resistance
        return turn_resistance

    def __str__(self):

        return self.polygon.__str__()


class BottomTurn:
    """Defines a bottom layer turn of a CFFC inductor
    """

    def __init__(
        self,
        at,
        inner_radius: float,
        outer_radius: float,
        gap: float,
        termination_width: float,
        viastrip_angle: float,
        viastrip_width: float,
        layer: str,
    ):

        self.inner_radius = inner_radius
        self.outer_radius = outer_radius
        self.layer = layer

        # calculate the gap angles
        inner_gap_angle = math.asin(gap / inner_radius)
        outer_gap_angle = math.asin(gap / outer_radius)

        # angle from "at" to the corner of the termination
        term_angle = math.asin(termination_width / outer_radius / 2)

        termination_arc = Arc(at, inner_radius, term_angle, -term_angle)

        # create the inner arc
        inner_arc = Arc(
            at,
            inner_radius + viastrip_width,
            2 * math.pi - term_angle,
            term_angle + inner_gap_angle + viastrip_angle,
        )

        via_arc = Arc(
            at,
            inner_radius,
            term_angle + inner_gap_angle + viastrip_angle,
            term_angle + inner_gap_angle,
        )

        # create outer arc
        outer_arc = Arc(
            at, outer_radius, term_angle + outer_gap_angle, 2 * math.pi - term_angle,
        )

        # create termination
        termination = [
            at + Point(outer_radius + termination_width, -termination_width / 2),
            at + Point(outer_radius + termination_width, termination_width / 2),
            at + Point(outer_radius * math.cos(term_angle), termination_width / 2),
        ]

        # create the polygon
        points = [termination_arc, inner_arc, via_arc, outer_arc] + termination
        self.polygon = Polygon(points, layer)

    def estimate_dcr(self, thickness: float, rho: float = 1.68e-8):
        """Estimate the DC resistance of the winding

        This function will estimate the DC resistance of the winding by calculating the estimated
        dc resistance of each turn and adding the estimated inter-turn via resistance 
        
        Args:
            thickness (float): The copper thickness of the layer
            rho (float): The conductivity of the material used in the layer

        Returns:
            float: An estimation of the DC resistance in ohms
        """

        # estimate the resistance of the turn
        turn_resistance = dcr_of_annulus(
            thickness, self.inner_radius, self.outer_radius, rho
        )

        # TODO:  Need to add via resistance
        return turn_resistance

    def __str__(self):

        return self.polygon.__str__()


class InnerTurn:
    """Defines a middle layer turn of a CFFC inductor
    """

    def __init__(
        self,
        at,
        inner_radius: float,
        outer_radius: float,
        gap: float,
        rotation: float,
        viastrip_angle: float,
        viastrip_width: float,
        layer: str,
    ):
        assert (
            outer_radius > inner_radius
        ), "outer radius must be greater than inner radius"

        self.inner_radius = inner_radius
        self.outer_radius = outer_radius
        self.layer = layer

        # calculate the gap angles
        inner_gap_angle = math.asin(gap / inner_radius)
        outer_gap_angle = math.asin(gap / outer_radius)

        # create the arcs
        start_via_arc = Arc(
            at,
            inner_radius,
            inner_gap_angle / 2 + rotation,
            inner_gap_angle / 2 + viastrip_angle + rotation,
        )
        inner_arc = Arc(
            at,
            inner_radius + viastrip_width,
            inner_gap_angle / 2 + viastrip_angle + rotation,
            2 * math.pi - inner_gap_angle / 2 - viastrip_angle + rotation,
        )
        end_via_arc = Arc(
            at,
            inner_radius,
            2 * math.pi - inner_gap_angle / 2 - viastrip_angle + rotation,
            2 * math.pi - inner_gap_angle / 2 + rotation,
        )
        outer_arc = Arc(
            at,
            outer_radius,
            2 * math.pi - outer_gap_angle / 2 + rotation,
            outer_gap_angle / 2 + rotation,
        )

        # create the polygon
        points = [start_via_arc, inner_arc, end_via_arc, outer_arc]
        self.polygon = Polygon(points, layer)

    def estimate_dcr(self, thickness: float, rho: float = 1.68e-8):
        """Estimate the DC resistance of the winding

        This function will estimate the DC resistance of the winding by calculating the estimated
        dc resistance of each turn and adding the estimated inter-turn via resistance 
        
        Args:
            thickness (float): The copper thickness of the layer
            rho (float): The conductivity of the material used in the layer

        Returns:
            float: An estimation of the DC resistance in ohms
        """

        # estimate the resistance of the turn
        turn_resistance = dcr_of_annulus(
            thickness, self.inner_radius, self.outer_radius, rho
        )

        # TODO:  Need to add via resistance
        return turn_resistance

    def __str__(self):
        return self.polygon.__str__()


class ViaStrip:
    def __init__(
        self,
        at: Point,
        layers: (str),
        inner_radius: float,
        start_angle: float,
        end_angle: float,
        size: float = 0.8e-3,
        drill: float = 0.4e-3,
    ):

        min_spacing = 0.5e-3

        # calculate how may vias we can fit in the strip
        angle = end_angle - start_angle
        width = inner_radius * abs(angle)
        number_vias = int(width / (drill + min_spacing))

        # calculate via locations
        via_radius = inner_radius + size / 2
        delta_angle = angle / number_vias
        initial_angle = start_angle + delta_angle / 2
        angles = [initial_angle + n * delta_angle for n in range(number_vias)]
        locations = [
            Point(via_radius * math.cos(angle), via_radius * math.sin(angle))
            for angle in angles
        ]

        # create via strip
        self.vias = [Via(location + at, size, drill, layers) for location in locations]

    def __str__(self):
        expression = "\n".join([via.__str__() for via in self.vias])
        return expression


if __name__ == "__main__":
    turn = InnerTurn(Point(0, 0), 6e-3, 12e-3, 0.5e-3, 0, math.pi / 8, 0, "F.Cu")
    print(turn)



from __future__ import annotations
import math

from planar_magnetics.geometry import (
    Point,
    Arc,
    Polygon,
    get_distance,
    TWO_PI,
    PI_OVER_TWO,
    THREE_PI_OVER_TWO,
)


def get_quadrant(angle: float):
    """Calculate the quadrant of an angle"""

    angle %= TWO_PI

    if angle < 0:
        angle += TWO_PI

    if angle < PI_OVER_TWO:
        return 1
    if angle < math.pi:
        return 2
    if angle < THREE_PI_OVER_TWO:
        return 3
    return 4


def smooth_point_to_arc(point: Point, arc: Arc, radius: float):
    """Find smoothing arc that joins section of polygon formed from a point and an arc

    Assuming you have a section of polygon formed from a line segment to an arc, this finds the arc
    that can be inserted between the line segment and the arc which is tangential to each

    Args:
        p (Point): The point
        arc (Arc): The Arc

    Returns:
        (Arc): The smoothing corner arc
    """

    # normalized to the arc center, which simplifies the math.  We will add the arc center back to
    # the result at the end
    p1 = point - arc.center
    p2 = arc.start - arc.center

    # calculate the angle of the vector p1 to p2
    segment_angle = math.atan2(p2.y - p1.y, p2.x - p1.x)

    # calculate the initial angle that the arc is pointing
    if arc.rotates_clockwise():
        arc_angle = arc.start_angle - PI_OVER_TWO
    else:
        arc_angle = arc.start_angle + PI_OVER_TWO

    # calculate the orientation of the corner relative to the line segment
    # and use to determine how to calculate the delta for the distance arithmetic bellow
    if get_quadrant(arc_angle - segment_angle + PI_OVER_TWO) > 2:
        delta = p2 - p1  # positive orientation
        positive_orientation = True
    else:
        delta = p1 - p2  # negative orientation
        positive_orientation = False

    # calculate the amplitude of the line segment
    R = abs(delta)

    # calculate the distance from the center of the arc to the center of the corner
    # if the segment intersects from the inside of the circle, it will be arc.radius - radius
    # if the segment intersects from the outside of the circle, it will be arc.radius + radius
    if abs(point - arc.center) < arc.radius:
        center_to_center = arc.radius - radius
        corner_inside_arc = True
    else:
        center_to_center = arc.radius + radius
        corner_inside_arc = False

    # calculate the angle of the vector from center to center
    alpha = math.atan2(delta.y, -delta.x)

    if arc.rotates_clockwise():
        angle = (
            math.pi
            - math.asin(
                (radius * R - delta.x * p1.y + delta.y * p1.x) / center_to_center / R
            )
            - alpha
        )
    else:
        angle = (
            math.asin(
                (radius * R - delta.x * p1.y + delta.y * p1.x) / center_to_center / R
            )
            - alpha
        )

    # derive center of the corner
    x = center_to_center * math.cos(angle)
    y = center_to_center * math.sin(angle)
    center = Point(x, y) + arc.center

    # derive the start and end angles
    start_angle = PI_OVER_TWO + math.atan2(delta.y, delta.x)

    if corner_inside_arc:
        end_angle = angle
    else:
        end_angle = angle + math.pi

    # make sure arc rotates in correct direction
    # positive orientations need to rotate clockwise
    if positive_orientation:
        while start_angle < end_angle:
            start_angle += TWO_PI
    else:
        while start_angle > end_angle:
            start_angle -= TWO_PI

    return Arc(center, radius, start_angle, end_angle)


def round_corner(arc1: Arc, arc2: Arc, radius: float):

    # check if the transition from the end of arc1 is continuous
    if math.isclose(arc1.radius, get_distance(arc1.center, arc1.end, arc2.start)):
        # if so, we just add the original arc
        arcs = [arc1]
    else:
        # otherwise we add a smoothing corner
        corner = smooth_point_to_arc(arc2.start, arc1.reverse(), radius).reverse()
        arcs = [
            Arc(arc1.center, arc1.radius, arc1.start_angle, corner.start_angle),
            corner,
        ]
    # check if the transition to the start of arc2 is continuous
    if math.isclose(arc2.radius, get_distance(arc2.center, arc1.end, arc2.start)):
        # if so, we just add the original arc
        return arcs + [arc2]

    # otherwise we add a smoothing corner
    corner = smooth_point_to_arc(arc1.end, arc2, radius)

    if abs(corner.center - arc2.center) < arc2.radius:
        start_angle = corner.end_angle
    else:
        start_angle = corner.end_angle - math.pi

    arcs += [
        corner,
        Arc(arc2.center, arc2.radius, start_angle, arc2.end_angle),
    ]
    return arcs


def smooth_polygon(polygon: Polygon, radius: float) -> Polygon:
    """Smooth the corners of a polygon

    Adds smooth transitions with tangential arcs in between the points of a polygon

    Args:
        polygon (Polygon): The original polygon to smooth
        radius (float): The radius of the smoothing arcs

    Returns:
        Polygon: A new polygon with smoothed corners
    """

    assert radius > 0, "The radius must be a positive number"

    # smooth the corners
    arcs = [polygon.points[0]]
    for arc in polygon.points[1:]:
        arcs.extend(round_corner(arcs.pop(), arc, radius))

    # don't forget to smooth the start-to-finish transition
    arcs.extend(round_corner(arcs.pop(), arcs.pop(0), radius))

    return Polygon(arcs, polygon.layer, polygon.width, polygon.fill)



from __future__ import annotations
from typing import Union
from pathlib import Path
from dataclasses import dataclass, field
import math
import uuid

# useful geometric constants
TWO_PI = 2 * math.pi
PI_OVER_TWO = math.pi / 2
THREE_PI_OVER_TWO = 3 * math.pi / 2


def get_oriented_distance(p0: Point, p1: Point, p2: Point):
    """Calculate the oriented distance between a point and a line

    Calculate the distance between a point p0 and a line formed between p1 and p2.  This result is
    the "oriented" distance, meaning it is signed.  What this means that if you thought of the line
    from p1-to-p2 as a vector that pointed up, as positive result would mean p0 was on the right
    side of the vector and a negative result would mean p0 was on the left side.
    """
    p21 = p2 - p1
    p10 = p1 - p0

    return (p21.x * p10.y - p21.y * p10.x) / abs(p21)


def get_distance(p0: Point, p1: Point, p2: Point):
    """Calculate the distance between a point and a line

    Calculate the distance between a point p0 and a line formed between p1 and p2
    """

    return abs(get_oriented_distance(p0, p1, p2))


@dataclass
class Point:
    x: float
    y: float

    def __str__(self):
        return f"{self.x*1e3} {self.y*1e3}"

    def __add__(self, other: Point) -> Point:
        return Point(self.x + other.x, self.y + other.y)

    def __sub__(self, other: Point) -> Point:
        return Point(self.x - other.x, self.y - other.y)

    def __abs__(self) -> float:
        return math.sqrt(self.x ** 2 + self.y ** 2)

    def __eq__(self, other: Point) -> bool:
        if not math.isclose(self.x, other.x):
            return False
        if not math.isclose(self.y, other.y):
            return False
        return True

    def __ne__(self, other: Point) -> bool:
        return not self.__eq__(other)


def point_from_polar(radius, angle):
    x = radius * math.cos(angle)
    y = radius * math.sin(angle)
    return Point(x, y)


@dataclass
class Arc:
    center: Point
    radius: float
    start_angle: float
    end_angle: float
    start: Point = field(init=False)
    mid: Point = field(init=False)
    end: Point = field(init=False)
    bulge: float = field(init=False)

    def __post_init__(self):
        """Derived parameters"""

        # three-point representation
        mid_angle = (self.start_angle + self.end_angle) / 2
        self.start = self.center + point_from_polar(self.radius, self.start_angle)
        self.mid = self.center + point_from_polar(self.radius, mid_angle)
        self.end = self.center + point_from_polar(self.radius, self.end_angle)

        # bulge (for dxf generation)
        width = abs(self.end - self.start)
        sagitta = get_oriented_distance(self.mid, self.start, self.end)
        self.bulge = 2 * sagitta / width

    def __str__(self):
        return f"(arc (start {self.start}) (mid {self.mid}) (end {self.end}))"

    def __add__(self, other: Point):
        return Arc(self.at + other, self.radius, self.start_angle, self.end_angle)

    def rotates_clockwise(self):
        return self.end_angle < self.start_angle

    def rotates_counterclockwise(self):
        return self.end_angle > self.start_angle

    def reverse(self):
        return Arc(self.center, self.radius, self.end_angle, self.start_angle)

    def rotate(self, angle: float):
        """Rotate an arc about its center"""
        return Arc(
            self.center, self.radius, self.start_angle + angle, self.end_angle + angle
        )

    def rotate_about(self, about: Point, angle: float):
        """ Rotate an arc around a reference point"""

        # calculate the new center for the arc
        offset = self.center - about
        sin_angle = math.sin(angle)
        cos_angle = math.cos(angle)
        x = offset.x * cos_angle - offset.y * sin_angle
        y = offset.y * cos_angle + offset.x * sin_angle
        center = about + Point(x, y)

        return Arc(
            center, self.radius, self.start_angle + angle, self.end_angle + angle
        )

    def interpolate(self, max_angle: float = math.pi / 36):
        """Create a PWL approximation of the arc with a list of points
        """
        points = []
        angle = self.start_angle
        if self.rotates_counterclockwise():
            while angle < self.end_angle:
                point = Point(
                    self.radius * math.cos(angle), self.radius * math.sin(angle)
                )
                points.append(self.center + point)
                angle += max_angle
        else:
            while angle > self.end_angle:
                point = Point(
                    self.radius * math.cos(angle), self.radius * math.sin(angle)
                )
                points.append(self.center + point)
                angle -= max_angle

        point = Point(
            self.radius * math.cos(self.end_angle),
            self.radius * math.sin(self.end_angle),
        )
        points.append(self.center + point)
        return points

    def add_to_dxf_model(self, modelspace):
        """Add Arc to DXF model"""

        modelspace.add_arc(
            center=(self.center.x, self.center.y),
            radius=self.radius,
            start_angle=self.start_angle,
            end_angle=self.end_angle,
        )


@dataclass
class Polygon:
    points: [Point]
    layer: str = "F.Cu"
    width: float = 0
    fill: str = "solid"
    tstamp: uuid.UUID = uuid.uuid4()

    def __add__(self, other: Point):
        return Polygon(
            [point + other for point in self.points], self.layer, self.width, self.fill
        )

    def __str__(self):
        points = "".join(
            [
                f"{point}" if isinstance(point, Arc) else f"(xy {point})"
                for point in self.points
            ]
        )
        expression = f"(fp_poly(pts{points})(layer {self.layer}) (width {self.width}) (fill {self.fill}) (tstamp {self.tstamp}))"
        return expression

    def to_poly_path(self):
        """Create a true arc polypath
        """
        points = []
        last = None
        for point in self.points:

            if isinstance(point, Arc):
                # if the start of this is not the end of the last arc, we need to add a point
                if last is not None and point.start != last:
                    points.append((last.x, last.y))

                points.append((point.start.x, point.start.y, 0, 0, point.bulge))
                last = point.end
                continue

            if last is not None and point != last:
                points.append((last.x, last.y))
            points.append((point.x, point.y))
            last = None

        # Make sure to add the last point if there is one left over from an arc
        if last is not None:
            points.append((last.x, last.y))

        return points

    def to_pwl_path(self, max_angle: float = math.pi / 36):
        """Create a list of tuples representing the polypath as a PWL approximation
        """
        points = []
        for point in self.points:

            if isinstance(point, Arc):
                points.extend([(p.x, p.y) for p in point.interpolate(max_angle)])
                continue
            points.append((point.x, point.y))
        return points

    def export_to_dxf(
        self,
        filename: Union[str, Path],
        version: str = "R2000",
        encoding: str = None,
        fmt: str = "asc",
    ) -> None:
        """Export the polygon to a dxf file

        Args:
            filename: file name as string
            version: DXF version
            encoding: override default encoding as Python encoding string like ``'utf-8'``
            fmt: ``'asc'`` for ASCII DXF (default) or ``'bin'`` for Binary DXF
        """
        import ezdxf

        doc = ezdxf.new(version)
        msp = doc.modelspace()
        path = self.to_poly_path()
        msp.add_lwpolyline(path, close=True)
        doc.saveas(filename, encoding, fmt)

    def plot(self, max_angle: float = math.pi / 36):
        """Create a plot preview of the polygon
        """
        import matplotlib.pyplot as mp

        path = self.to_pwl_path(max_angle)
        mp.figure(figsize=(8, 8))
        mp.axis("equal")
        x, y = zip(*path)
        mp.fill(x, y)
        mp.show()


if __name__ == "__main__":

    p0 = Point(0, 4)
    p1 = Point(4, 4)
    p2 = Point(4, 0)
    p3 = Point(0, 0)
    arc = Arc(Point(2, 4), 1, math.pi, 0)
    polygon = Polygon([p0, arc, p1, p2, p3])
    # polygon.plot(max_angle=math.pi / 8)

    path = polygon.to_poly_path()
    print(path)



from __future__ import annotations
from planar_magnetics.geometry import Point
from dataclasses import dataclass
import uuid
from enum import Enum


class PadType(Enum):
    TH = 1
    SMD = 2

    def __str__(self):
        if self.value == 1:
            return "thru_hole"
        else:
            return "smd"


@dataclass
class Pad:
    pad_type: PadType
    number: int
    at: Point
    size: float
    layers: (str) = ("*.Cu",)
    drill: float = None
    tstamp: uuid.UUID = uuid.uuid4()

    def __str__(self):

        layers = " ".join(self.layers)
        drill_expression = f"(drill {self.drill*1e3})" if self.drill else ""
        shape = "circle" if self.pad_type == PadType.TH else "rect"
        expression = f'(pad "{self.number}" {self.pad_type} {shape} (at {self.at}) (size {self.size*1e3} {self.size*1e3}) {drill_expression} (layers {layers}) (remove_unused_layers) (tstamp {self.tstamp}))'
        return expression


@dataclass
class Via:
    at: Point = Point(0, 0)
    size: float = 0.8
    drill: float = 0.4
    layers: (str) = ("F.Cu",)
    remove_unused_layers = True
    tstamp: uuid.UUID = uuid.uuid4()

    def to_pad(self, number: int = 1):
        """Convert to an equivalent through-hole pad"""

        return Pad(PadType.TH, number, self.at, self.size, ("*.Cu",), self.drill)

    def __str__(self):
        layers = " ".join(self.layers)
        annular_option = "(remove_unused_layers)" if self.remove_unused_layers else ""
        return f"(via (at {self.at}) (size {self.size*1e3}) (drill {self.drill*1e3}) (layers {layers}) {annular_option} (free) (net 0) (tstamp {self.tstamp}))"


@dataclass
class Reference:
    at: Point
    font_size: float = 1.27e-3
    thickness: float = 0.15e-3
    justification: str = "left"
    tstamp: uuid.UUID = uuid.uuid4()

    def __str__(self):

        expression = f'(fp_text reference "Ref**" (at {self.at}) (layer "F.SilkS") (effects (font (size {self.font_size*1e3} {self.font_size*1e3}) (thickness {self.thickness*1e3})) (justify {self.justification})) (tstamp {self.tstamp}))'

        return expression


@dataclass
class Value:
    at: Point
    font_size: float = 1.27e-3
    thickness: float = 0.15e-3
    justification: str = "left"
    tstamp: uuid.UUID = uuid.uuid4()

    def __str__(self):

        expression = f'(fp_text value "Val**" (at {self.at}) (layer "F.SilkS") (effects (font (size {self.font_size*1e3} {self.font_size*1e3}) (thickness {self.thickness*1e3})) (justify {self.justification})) (tstamp {self.tstamp}))'

        return expression


class Footprint:
    """Python representation of a KiCAD footprint
    """

    def __init__(
        self, name: str, version: str = "20211014", contents=[],
    ):
        self.name = name
        self.version = version
        self.contents = contents

    def __str__(self):
        header = f'footprint "{self.name}" (version {self.version}) (generator python_planar_magnetics)'
        contents = "\n".join([content.__str__() for content in self.contents])
        expression = f"({header} {contents})"

        return expression



import math


def weight_to_thickness(weight: float):
    """Converter a copper weight to a thickness

    Converters a copper weight (oz) to a thickess (meters)

    Args:
        weight: weight of copper layer in oz

    Returns:
        float: The copper thickness in meters
    """

    return 35e-6 * weight


def dcr_of_sheet(thickness: float, width: float, length: float, rho: float = 1.68e-8):
    """Calculate the resistance of a rectangular sheet
    """
    resistance = (length * rho) / (width * thickness)
    return resistance


def dcr_of_annulus(
    thickness: float, inner_radius: float, outer_radius: float, rho: float = 1.68e-8
):
    """Calculate the sheet resistance of an annulus
    """

    assert (
        outer_radius > inner_radius
    ), f"Outer radius must be greater than inner radius"

    resistance = (2 * math.pi * rho) / (
        thickness * (math.log(outer_radius / inner_radius))
    )

    return resistance



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="planar-magnetics",
    version="0.1.2",
    author="Donny Zimmanck",
    author_email="dzimmanck@gmail.com",
    description="Create planar magnetic structures programmatically and export to CAD tools.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dzimmanck/python-planar-magnetics",
    project_urls={
        "Bug Tracker": "https://github.com/dzimmanck/python-planar-magnetics/issues",
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(),
    install_requires=["ezdxf>=0.17.2"],
    extras_require={"dev": ["pytest",],},
    python_requires=">=3.7",
)
